# -*- coding: utf-8 -*-

"""
utils.py: collection of utilities for testing the mf2 package
"""


from collections import namedtuple
import numpy as np


ValueRange = namedtuple('ValueRange', ['min', 'max'])

def determinerange(values):
    """Determine the range of values in each dimension"""
    r = ValueRange(np.min(values, axis=0), np.max(values, axis=0))
    if np.any(r.max - r.min < 1e-8):
        r = ValueRange(r.min - 1e8, r.max + 1e8)
    return r


def rescale(values, *, range_in=None, range_out=ValueRange(0, 1), scale_only=False):
    """Perform a scale transformation of `values`: [range_in] --> [range_out]"""

    if range_in is None:
        range_in = determinerange(values)
    elif not isinstance(range_in, ValueRange):
        range_in = ValueRange(*range_in)

    if not isinstance(range_out, ValueRange):
        range_out = ValueRange(*range_out)

    scale_out = range_out.max - range_out.min
    scale_in = range_in.max - range_in.min

    if scale_only:
        scaled_values = (values / scale_in) * scale_out
    else:
        scaled_values = (values - range_in.min) / scale_in
        scaled_values = (scaled_values * scale_out) + range_out.min

    return scaled_values



# -*- coding: utf-8 -*-

"""Implementation of the adjustable bi-fidelity Hartmann3 function
as defined in:

    Toal, D.J.J. Some considerations regarding the use of multi-
    fidelity Kriging in the construction of surrogate models.
    Struct Multidisc Optim 51, 1223–1245 (2015)
    doi:10.1007/s00158-014-1209-5
"""


from functools import partial

import numpy as np

from mf2.multiFidelityFunction import MultiFidelityFunction


# Some constant values
# Hartmann 3d
_alpha3 = np.array([1.0, 1.2, 3.0, 3.2])[:,np.newaxis]
_beta3 = np.array([
    [3.0, 10.0, 30.0],
    [0.1, 10.0, 35.0],
    [3.0, 10.0, 30.0],
    [0.1, 10.0, 35.0],
]).T[np.newaxis,:,:]
_P3 = np.array([
    [.3689, .1170, .2673],
    [.4699, .4387, .7470],
    [.1091, .8732, .5547],
    [.0381, .5743, .8828],
]).T[np.newaxis,:,:]


def hartmann3_hf(xx):
    xx = np.atleast_2d(xx)

    xx = xx[:,:,np.newaxis]

    tmp1 = (xx - _P3) ** 2 * _beta3
    tmp2 = np.exp(-np.sum(tmp1, axis=1))
    tmp3 = tmp2.dot(_alpha3)

    return -tmp3.reshape((-1,))


def adjustable_hartmann3_lf(xx, a3):
    xx = np.atleast_2d(xx)

    factor = 3/4 * (a3+1)

    xx = xx[:,:,np.newaxis]

    tmp1 = (xx - (_P3 * factor))
    tmp2 = tmp1 ** 2 * _beta3
    tmp3 = np.exp(-np.sum(tmp2, axis=1))
    tmp4 = tmp3.dot(_alpha3)

    return -tmp4.reshape((-1,))


u_bound = [1]*3
l_bound = [0]*3


def hartmann3(a3: float):
    """Factory method for adjustable Hartmann3 function using parameter value `a3`

    :param a3:  Parameter to tune the correlation between high- and low-fidelity
                functions. Expected values lie in range [0, 1]. High- and low-
                fidelity are identical for `a3=1/3`.
    :return:    A MultiFidelityFunction instance
    """

    return MultiFidelityFunction(
        f"adjustable Hartmann3 {a3}",
        u_bound, l_bound,
        [hartmann3_hf, partial(adjustable_hartmann3_lf, a3=a3)],
        fidelity_names=['high', 'low'],
    )




# -*- coding: utf-8 -*-

"""Implementation of the bi-fidelity Bohachevsky function as defined in:

    Dong, H., Song, B., Wang, P. et al. Multi-fidelity information
    fusion based on prediction of kriging. Struct Multidisc Optim
    51, 1267–1280 (2015) doi:10.1007/s00158-014-1213-9
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction


def bohachevsky_hf(xx):
    """
    BOHACHEVSKY FUNCTION

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = x1**2 + 2*x2**2
    term2 = 0.3*np.cos(3*np.pi*x1)
    term3 = 0.4*np.cos(4*np.pi*x2)

    return term1 - term2 - term3 + 0.7


def bohachevsky_lf(xx):
    """
    BOHACHEVSKY FUNCTION, LOWER FIDELITY CODE
    Calls: bohachevsky_hf
    This function, from Dong et al. (2015), is used as the "low-accuracy code"
    version of the function bohachevsky_hf.

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = bohachevsky_hf(np.hstack([0.7*x1.reshape(-1,1), x2.reshape(-1,1)]))
    term2 = x1*x2 - 12

    return term1 + term2


#: Lower bound for Bohachevsky function
l_bound = [-5, -5]
#: Upper bound for Bohachevsky function
u_bound = [ 5,  5]

#: 2D Bohachevsky function with fidelities 'high' and 'low'
bohachevsky = MultiFidelityFunction(
    "bohachevsky",
    u_bound, l_bound,
    [bohachevsky_hf, bohachevsky_lf],
    fidelity_names=['high', 'low']
)



# -*- coding: utf-8 -*-

"""Implementation of the adjustable bi-fidelity Branin function
as defined in:

    Toal, D.J.J. Some considerations regarding the use of multi-
    fidelity Kriging in the construction of surrogate models.
    Struct Multidisc Optim 51, 1223–1245 (2015)
    doi:10.1007/s00158-014-1209-5
"""


from functools import partial

import numpy as np

from mf2.multiFidelityFunction import MultiFidelityFunction
from mf2.branin import branin_base, l_bound, u_bound


_four_pi_square = 4*np.pi**2


def adjustable_branin_lf(xx, a1):
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = branin_base(xx)
    term2 = x2 - (5.1 * (x1**2 / _four_pi_square)) + ((5*x1) / np.pi) - 6

    return term1 - (a1+0.5) * term2**2


def branin(a1: float):
    """Factory method for adjustable Branin function using parameter value `a1`

    :param a1:  Parameter to tune the correlation between high- and low-fidelity
                functions. Expected values lie in range [0, 1]. High- and low-
                fidelity are identical for `a1=-0.5`.
    :return:    A MultiFidelityFunction instance
    """

    return MultiFidelityFunction(
        f"adjustable Branin {a1}",
        u_bound, l_bound,
        [branin_base, partial(adjustable_branin_lf, a1=a1)],
        fidelity_names=['high', 'low'],
    )



# -*- coding: utf-8 -*-

"""Implementation of the adjustable bi-fidelity Paciorek function
as defined in:

    Toal, D.J.J. Some considerations regarding the use of multi-
    fidelity Kriging in the construction of surrogate models.
    Struct Multidisc Optim 51, 1223–1245 (2015)
    doi:10.1007/s00158-014-1209-5
"""


from functools import partial

import numpy as np

from mf2.multiFidelityFunction import MultiFidelityFunction


def paciorek_hf(xx):
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T
    return np.sin(1/(x1*x2))


def adjustable_paciorek_lf(xx, a2):
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T
    temp1 = paciorek_hf(xx)
    temp2 = 9 * a2**2
    temp3 = np.cos(1/(x1*x2))
    return temp1 - (temp2*temp3)


u_bound = [1]*2
l_bound = [0.3]*2


def paciorek(a2: float):
    """Factory method for adjustable Paciorek function using parameter value `a2`

    :param a2:  Parameter to tune the correlation between high- and low-fidelity
                functions. Expected values lie in range [0, 1]. High- and low-
                fidelity are identical for `a2=0.0`.
    :return:    A MultiFidelityFunction instance
    """

    return MultiFidelityFunction(
        f"adjustable Paciorek {a2}",
        u_bound, l_bound,
        [paciorek_hf, partial(adjustable_paciorek_lf, a2=a2)],
        fidelity_names=['high', 'low'],
    )



# -*- coding: utf-8 -*-

"""
hartmann.py: contains the Hartmann6 function

As defined in

    "Remarks on multi-fidelity surrogates" by Chanyoung Park, Raphael T. Haftka
    and Nam H. Kim (2016)
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction

# Some constant values for the Hartmann 6d calculations
_alpha6_low = np.array([0.5, 0.5, 2.0, 4.0])[:,np.newaxis]
_alpha6_high = np.array([1.0, 1.2, 3.0, 3.2])[:, np.newaxis]
_A6 = np.array([
    [10.00,  3.0, 17.00,  3.5,  1.7,  8],
    [ 0.05, 10.0, 17.00,  0.1,  8.0, 14],
    [ 3.00,  3.5,  1.70, 10.0, 17.0,  8],
    [17.00,  8.0,  0.05, 10.0,  0.1, 14],
]).T[np.newaxis,:,:]
_P6 = np.array([
    [.1312, .1696, .5569, .0124, .8283, .5886],
    [.2329, .4135, .8307, .3736, .1004, .9991],
    [.2348, .1451, .3522, .2883, .3047, .6650],
    [.4047, .8828, .8732, .5743, .1091, .0381],
]).T[np.newaxis,:,:]
_four_nine_exp = np.exp(-4 / 9)


def hartmann6_hf(xx):
    xx = np.atleast_2d(xx)

    xx = xx[:,:,np.newaxis]

    tmp1 = (xx - _P6) ** 2 * _A6
    tmp2 = np.exp(-np.sum(tmp1, axis=1))
    tmp3 = tmp2.dot(_alpha6_high) + 2.58

    return -(1/1.94) * tmp3.reshape((-1,))


def hartmann6_lf(xx):
    xx = np.atleast_2d(xx)

    xx = xx[:,:,np.newaxis]

    tmp1 = (xx - _P6) ** 2 * _A6
    tmp2 = _f_exp(-np.sum(tmp1, axis=1))
    tmp3 = tmp2.dot(_alpha6_low) + 2.58

    return -(1/1.94) * tmp3.reshape((-1,))


def _f_exp(xx):
    return (_four_nine_exp + (_four_nine_exp * (xx + 4)/9)) ** 9


#: Lower bound for Hartmann6 function
l_bound = [0.1] * 6
#: Upper bound for Hartmann6 function
u_bound = [1] * 6

#: 6D Hartmann6 function with fidelities 'high' and 'low'
hartmann6 = MultiFidelityFunction(
    "Hartmann6",
    u_bound, l_bound,
    [hartmann6_hf, hartmann6_lf],
    fidelity_names=['high', 'low']
)



# -*- coding: utf-8 -*-

"""
adjustable/: correlation-adjustable multi-fidelity functions.
"""

__author__ = 'Sander van Rijn'
__email__ = 's.j.van.rijn@liacs.leidenuniv.nl'


from .branin import branin
from .hartmann import hartmann3
from .paciorek import paciorek
from .trid import trid


bi_fidelity_functions = (
    branin,
    paciorek,
    hartmann3,
    trid,
)



# -*- coding: utf-8 -*-

"""Implementation of the bi-fidelity Branin function as defined in:

    Dong, H., Song, B., Wang, P. et al. Multi-fidelity information
    fusion based on prediction of kriging. Struct Multidisc Optim
    51, 1267–1280 (2015) doi:10.1007/s00158-014-1213-9
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction


_four_pi_square = 4*np.pi**2
_eight_pi = 8*np.pi


def branin_base(xx):
    """
    BRANIN FUNCTION

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = x2 - (5.1 * (x1**2 / _four_pi_square)) + ((5*x1) / np.pi) - 6
    term2 = (10 * np.cos(x1)) * (1 - (1/_eight_pi))

    return term1**2 + term2 + 10


def branin_hf(xx):
    """
    BRANIN FUNCTION, HIGH FIDELITY CODE
    Calls: branin_base
    This function, from Dong et al. (2015), is used as the "high-accuracy code"
    version of the function based on the 'traditional' branin function.

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    _, x2 = xx.T
    return branin_base(xx) - 22.5*x2


def branin_lf(xx):
    """
    BRANIN FUNCTION, LOWER FIDELITY CODE
    Calls: branin_base
    This function, from Dong et al. (2015), is used as the "low-accuracy code"
    version of the function branin_hf.

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = branin_base(np.hstack([0.7*x1.reshape(-1,1), 0.7*x2.reshape(-1,1)]))
    term2 = 1.575*x2  # 15.75
    term3 = 2*(.9+x1**2)
    term4 = 50

    return term1 - term2 + term3 - term4


#: Lower bound for Branin function
l_bound = [-5,  0]
#: Upper bound for Branin function
u_bound = [10, 15]

#: 2D Branin function with fidelities 'high' and 'low'
branin = MultiFidelityFunction(
    "branin",
    u_bound, l_bound,
    [branin_hf, branin_lf],
    fidelity_names=['high', 'low'],
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Filename.py: << A short summary docstring about this file >>
"""

__author__ = 'Sander van Rijn'
__email__ = 's.j.van.rijn@liacs.leidenuniv.nl'


import numpy as np
import pytest

from .utils import rescale, ValueRange
import mf2


_functions_to_test = (
    mf2.forrester,
    mf2.Forrester(ndim=8),

    mf2.bohachevsky,
    mf2.booth,
    mf2.branin,
    mf2.currin,
    mf2.himmelblau,
    mf2.six_hump_camelback,
    mf2.park91a,
    mf2.park91b,
    mf2.hartmann6,
    mf2.borehole,

    mf2.adjustable.branin(0),
    mf2.adjustable.paciorek(0),
    mf2.adjustable.hartmann3(0),
    mf2.adjustable.trid(0),
)


def idfn(func):
    return f"{func.ndim}D-{func.name}"


@pytest.mark.parametrize("func", _functions_to_test, ids=idfn)
def test_function_regression(num_regression, func):
    np.random.seed(20160501)
    data_in = rescale(np.random.rand(100, func.ndim),
                      range_in=ValueRange(0,1),
                      range_out=ValueRange(*func.bounds))

    num_regression.check({
        fid: func[fid](data_in) for fid in func.fidelity_names
    })



from string import ascii_letters, printable

from hypothesis import given
from hypothesis.strategies import integers, lists, text
from mf2 import MultiFidelityFunction
from pytest import raises


@given(text(alphabet=printable, min_size=1, max_size=30))
def test_name(name):
    mff = MultiFidelityFunction(name, [0], [0], functions=None)
    assert mff._name == name
    assert mff.name == name.title()

def test_unequal_bound_lenghts():
    with raises(ValueError):
        bounds_A, bounds_B = [0], [1, 2]
        MultiFidelityFunction('test', bounds_A, bounds_B, functions=None)
        MultiFidelityFunction('test', bounds_B, bounds_A, functions=None)


@given(integers(0, 100))
def test_ndim_matches_bounds_length(ndim):
    u_bounds, l_bounds = [1]*ndim, [0]*ndim
    mff = MultiFidelityFunction('test', u_bounds, l_bounds, functions=None)
    assert mff.ndim == ndim


def _list_of_strings(n):
    return lists(text(alphabet=ascii_letters), min_size=1, max_size=n)


@given(_list_of_strings(100).filter(lambda x: len(x) == len(set(x))))
def test_access_with_fidelity_names(fidelity_names):
    functions = [lambda x: None for _ in fidelity_names]
    mff = MultiFidelityFunction(
        'test', [0], [1],
        functions=functions,
        fidelity_names=fidelity_names
    )

    for idx, name in enumerate(fidelity_names):
        assert mff[idx] is mff[name] is getattr(mff, name)


@given(integers(1, 100))
def test_access_without_fidelity_names(num_fidelities):
    mff = MultiFidelityFunction(
        'test', [0], [1],
        functions=[lambda idx=idx: idx for idx in range(num_fidelities+1)],
    )

    with raises(AttributeError):
        _ = mff.high()
        _ = mff.low()
    with raises(IndexError):
        _ = mff['high']()
        _ = mff['low']()

    for idx in range(num_fidelities):
        assert mff[idx]() == idx



# -*- coding: utf-8 -*-

"""Implementation of the bi-fidelity Booth function as defined in:

    Dong, H., Song, B., Wang, P. et al. Multi-fidelity information
    fusion based on prediction of kriging. Struct Multidisc Optim
    51, 1267–1280 (2015) doi:10.1007/s00158-014-1213-9
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction


def booth_hf(xx):
    """
    BOOTH FUNCTION

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = (x1 + 2*x2 - 7)**2
    term2 = (2*x1 + x2 - 5)**2

    return term1 + term2


def booth_lf(xx):
    """
    BOOTH FUNCTION, LOWER FIDELITY CODE
    Calls: booth_hf
    This function, from Dong et al. (2015), is used as the "low-accuracy code"
    version of the function booth_hf.

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = booth_hf(np.hstack([.4*x1.reshape(-1,1), x2.reshape(-1,1)]))
    term2 = 1.7*x1*x2 - x1 + 2*x2

    return term1 + term2


#: Lower bound for Booth function
l_bound = [-10, -10]
#: Upper bound for Booth function
u_bound = [ 10,  10]

#: 2D Booth function with fidelities 'high' and 'low'
booth = MultiFidelityFunction(
    "booth",
    u_bound, l_bound,
    [booth_hf, booth_lf],
    fidelity_names=['high', 'low']
)



# -*- coding: utf-8 -*-

"""
mf2

A collection of analytical functions with 2 or more available fidelities.
"""
import pkg_resources

from .multiFidelityFunction import MultiFidelityFunction, invert
from .borehole import borehole
from .currin import currin
from .park91a import park91a
from .park91b import park91b
from .bohachevsky import bohachevsky
from .branin import branin
from .booth import booth
from .forrester import forrester, Forrester
from .himmelblau import himmelblau
from .six_hump_camelback import six_hump_camelback
from .hartmann import hartmann6

import mf2.adjustable

__author__ = 'Sander van Rijn'
__email__ = 's.j.van.rijn@liacs.leidenuniv.nl'
__version__ = pkg_resources.get_distribution('mf2').version


bi_fidelity_functions = (
    # 1D
    forrester,
    # 2D
    bohachevsky,
    booth,
    branin,
    currin,
    himmelblau,
    six_hump_camelback,
    # 4D
    park91a,
    park91b,
    # 6D
    hartmann6,
    # 8D
    borehole,
)



# -*- coding: utf-8 -*-

"""Implementation of the bi-fidelity Himmelblau function as defined in:

    Dong, H., Song, B., Wang, P. et al. Multi-fidelity information
    fusion based on prediction of kriging. Struct Multidisc Optim
    51, 1267–1280 (2015) doi:10.1007/s00158-014-1213-9
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction


def himmelblau_hf(xx):
    """
    HIMMELBLAU FUNCTION

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = (x1**2 + x2 - 11)**2
    term2 = (x2**2 + x1 - 7)**2

    return term1 + term2


def himmelblau_lf(xx):
    """
    HIMMELBLAU FUNCTION, LOWER FIDELITY CODE
    Calls: himmelblau_hf
    This function, from Dong et al. (2015), is used as the "low-accuracy code"
    version of the function himmelblau_hf.

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = himmelblau_hf(np.hstack([0.5*x1.reshape(-1,1), 0.8*x2.reshape(-1,1)]))
    term2 = x2**3 - (x1 + 1)**2

    return term1 + term2


#: Lower bound for Himmelblau function
l_bound = [-4, -4]
#: Upper bound for Himmelblau function
u_bound = [ 4,  4]

#: 2D Himmelblau function with fidelities 'high' and 'low'
himmelblau = MultiFidelityFunction(
    "himmelblau",
    u_bound, l_bound,
    [himmelblau_hf, himmelblau_lf],
    fidelity_names=['high', 'low']
)



# -*- coding: utf-8 -*-

"""
forrester.py: Forrester function

This file contains the definition of an adapted version of the simple 1D
example function as presented in:

    Forrester Alexander I.J, Sóbester András and Keane Andy J "Multi-fidelity
    Optimization via Surrogate Modelling", Proceedings of the Royal Society A,
    vol. 463, http://doi.org/10.1098/rspa.2007.1900

This version has been adapted to be multi-dimensional, input can be arbitrarily
many dimensions. Output value is calculated as the mean of the outcomes for all
separate dimensions.
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction


def forrester_high(X):
    X = np.atleast_2d(X)

    ndim = X.shape[1]
    term1 = (6*X - 2)**2
    term2 = np.sin(12*X - 4)
    return np.sum(term1 * term2, axis=1) / ndim


def forrester_low(X, *, A=0.5, B=10, C=-5):
    X = np.atleast_2d(X)

    ndim = X.shape[1]
    term1 = A*forrester_high(X)
    term2 = B*(X - 0.5)
    term3 = C

    return term1 + (np.sum(term2, axis=1) / ndim) + term3


#: Lower bound for Forrester function
l_bound = [0]
#: Upper bound for Forrester function
u_bound = [1]


def Forrester(ndim: int):
    """Factory method for `ndim`-dimensional multi-fidelity Forrester function

    :param ndim: Desired dimensionality
    :return:     :class:`~mf2.multiFidelityFunction.MultiFidelityFunction`
                 instance with bounds of appropriate length
    """
    if ndim < 1:
        raise ValueError(f"ndim must be at least 1, not {ndim}")

    return MultiFidelityFunction(
        "forrester",
        u_bound=np.repeat(u_bound, ndim),
        l_bound=np.repeat(l_bound, ndim),
        functions=[forrester_high, forrester_low],
        fidelity_names=['high', 'low'],
    )


#: 1D Forrester function with fidelities 'high' and 'low'
forrester = Forrester(ndim=1)

#: 1D Forrester function with single fidelity 'high'
forrester_sf = MultiFidelityFunction(
    "forrester single fidelity",
    u_bound=u_bound, l_bound=l_bound,
    functions=[forrester_high],
    fidelity_names=['high'],
)



import setuptools

with open('README.md') as f:
    long_description = f.read()
with open('requirements.txt') as f:
    requirements = f.read().splitlines()
with open('requirements-dev.txt') as f:
    requirements_dev = f.read().splitlines()

setuptools.setup(
    name='mf2',
    version='2021.10.0',
    description='A collection of analytical benchmark functions in multiple fidelities',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Sander van Rijn',
    author_email='s.j.van.rijn@liacs.leidenuniv.nl',
    url="https://github.com/sjvrijn/mf2",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={"dev": requirements_dev},
)



# -*- coding: utf-8 -*-

"""Implementation of the bi-fidelity Six-hump Camel-back function
as defined in:

    Dong, H., Song, B., Wang, P. et al. Multi-fidelity information
    fusion based on prediction of kriging. Struct Multidisc Optim
    51, 1267–1280 (2015) doi:10.1007/s00158-014-1213-9
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction


def six_hump_camelback_hf(xx):
    """
    SIX-HUMP CAMEL-BACK FUNCTION

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T
    x1sq, x2sq = x1*x1, x2*x2

    term1 = (4 - 2.1*x1sq + (x1sq*x1sq)/3) * x1sq
    term2 = x1*x2
    term3 = (-4 + 4*x2sq) * x2sq

    return term1 + term2 + term3


def six_hump_camelback_lf(xx):
    """
    SIX-HUMP CAMEL-BACK FUNCTION, LOWER FIDELITY CODE
    Calls: sixHumpCamelBack_hf
    This function, from Dong et al. (2015), is used as the "low-accuracy code"
    version of the function sixHumpCamelBack_hf.

    INPUT:
    xx = [x1, x2]
    """
    xx = np.atleast_2d(xx)

    x1, x2 = xx.T

    term1 = six_hump_camelback_hf(np.hstack([0.7 * x1.reshape(-1, 1), 0.7 * x2.reshape(-1, 1)]))
    term2 = x1*x2 - 15

    return term1 + term2


#: Lower bound for Six-hump Camelback function
l_bound = [-2, -2]
#: upper bound for Six-hump Camelback function
u_bound = [ 2,  2]

#: 2D Six-hump Camelback function with fidelities 'high' and 'low'
six_hump_camelback = MultiFidelityFunction(
    "six hump camelback",
    u_bound, l_bound,
    [six_hump_camelback_hf, six_hump_camelback_lf],
    fidelity_names=['high', 'low']
)



# -*- coding: utf-8 -*-

"""Implementation of the bi-fidelity Park ('91) A function as defined in:

    Shifeng Xiong, Peter Z. G. Qian & C. F. Jeff Wu (2013) Sequential
    Design and Analysis of High-Accuracy and Low-Accuracy Computer
    Codes, Technometrics, 55:1, 37-46, DOI: 10.1080/00401706.2012.723572


Adapted from matlab implementation at

    https://www.sfu.ca/~ssurjano/park91a.html, retrieved 2017-10-02

by: Sonja Surjanovic and Derek Bingham, Simon Fraser University

Copyright 2013. Derek Bingham, Simon Fraser University.

THERE IS NO WARRANTY, EXPRESS OR IMPLIED. WE DO NOT ASSUME ANY LIABILITY
FOR THE USE OF THIS SOFTWARE.  If software is modified to produce
derivative works, such modified software should be clearly marked.
Additionally, this program is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; version 2.0 of the License.
Accordingly, this program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.
"""

import numpy as np

from .multiFidelityFunction import MultiFidelityFunction


def park91a_hf(xx):
    """
    PARK (1991) FUNCTION 1

    INPUT:
    xx = [x1, x2, x3, x4]
    """
    xx = np.atleast_2d(xx)

    x1, x2, x3, x4 = xx.T

    term1a = x1 / 2
    term1b = np.sqrt(1 + (x2 + x3 ** 2) * x4 / (x1 ** 2)) - 1
    term1 = term1a * term1b

    term2a = x1 + 3 * x4
    term2b = np.exp(1 + np.sin(x3))
    term2 = term2a * term2b

    return term1 + term2


def park91a_lf(xx):
    """
    PARK (1991) FUNCTION 1, LOWER FIDELITY CODE
    Calls: park91a_hf
    This function, from Xiong et al. (2013), is used as the "low-accuracy
    code" version of the function park91a_hf.

    INPUT:
    xx = [x1, x2, x3, x4]
    """
    xx = np.atleast_2d(xx)

    x1, x2, x3, _ = xx.T
    yh = park91a_hf(xx)

    term1 = (1 + np.sin(x1) / 10) * yh
    term2 = -2 * x1 + x2 ** 2 + x3 ** 2

    return term1 + term2 + 0.5


#: Lower bound for Park91A function
l_bound = [1e-8, 0, 0, 0]
#: Upper bound for Park91A function
u_bound = [1, 1, 1, 1]

#: 4D Park91A function with fidelities 'high' and 'low'
park91a = MultiFidelityFunction(
    "park91a",
    u_bound, l_bound,
    [park91a_hf, park91a_lf],
    fidelity_names=['high', 'low']
)



# -*- coding: utf-8 -*-

"""
multiFidelityFunction.py:

Defines the MultiFidelityFunction class for encapsuling all fidelities and
parameters of a multi-fidelity function. Also contains any other utility
functions that are commonly used by the various mf-functions in this package.
"""

from numbers import Integral
from typing import Callable
import numpy as np


class MultiFidelityFunction:

    def __init__(self, name, u_bound, l_bound, functions, fidelity_names=None):
        """All fidelity levels and parameters of a multi-fidelity function.

        :param name:           Name of the multi-fidelity function.
        :param u_bound:        Upper bound of the intended input range. Length
                               is also used to determine the (fixed)
                               dimensionality of the function.
        :param l_bound:        Lower bound of the intended input range. Must be
                               of same length as `u_bound`.
        :param functions:      Iterable of function handles for the different
                               fidelities, assumed to be sorted in *descending*
                               order.
        :param fidelity_names: List of names for the fidelities. Must be given
                               to support dictionary- or attribute-style
                               fidelity indexing, such as `f['high']()` and
                               `f.high()`
        """
        self._name = name

        if len(u_bound) != len(l_bound):
            raise ValueError(f"Length of upper and lower bounds are not equal: "
                             f"{len(u_bound)} != {len(l_bound)}")

        self.u_bound = np.array(u_bound, dtype=float)
        self.l_bound = np.array(l_bound, dtype=float)

        self.functions = functions
        if fidelity_names:
            # dict-style name-indexing
            self.fidelity_names = fidelity_names
            self.fidelity_dict = dict(zip(fidelity_names, functions))
            # class-style indexing
            for name, func in zip(fidelity_names, functions):
                setattr(self, name, func)
        else:
            self.fidelity_dict = None
            self.fidelity_names = None

    @property
    def name(self):
        return self._name.title()


    @property
    def ndim(self):
        """Dimensionality of the function. Inferred as ``len(self.u_bound)``."""
        return len(self.u_bound)


    @property
    def bounds(self):
        """Lower and upper bounds as a single np.array of shape (2, ndim)."""
        return np.array([self.l_bound, self.u_bound], dtype=float)


    def __getitem__(self, item):
        if isinstance(item, Integral):
            return self.functions[item]
        elif isinstance(item, str) and self.fidelity_dict:
            return self.fidelity_dict[item]
        else:
            raise IndexError(f"Invalid index '{item}'")


    def __repr__(self):
        return f"MultiFidelityFunction({self.name}, {self.u_bound}, {self.l_bound}, fidelity_names={self.fidelity_names})"


def invert(mff: MultiFidelityFunction) -> MultiFidelityFunction:
    """Invert a MultiFidelityFunction by multiplying all fidelities by -1

    :param mff: The MultiFidelityFunction to invert
    :return:     A new MultiFidelityFunction with the inverted fidelities
    """

    functions = [_invert_function(f) for f in mff.functions]

    return MultiFidelityFunction(
        mff._name, mff.u_bound, mff.l_bound,
        functions,
        fidelity_names=mff.fidelity_names,
    )


def _invert_function(func: Callable) -> Callable:
    """Applies a *-1 modification to the given function"""

    def inverted(x):
        return func(x) * -1

    return inverted



# -*- coding: utf-8 -*-

"""Implementation of the adjustable bi-fidelity Trid function
as defined in:

    Toal, D.J.J. Some considerations regarding the use of multi-
    fidelity Kriging in the construction of surrogate models.
    Struct Multidisc Optim 51, 1223–1245 (2015)
    doi:10.1007/s00158-014-1209-5
"""


from functools import partial

import numpy as np

from mf2.multiFidelityFunction import MultiFidelityFunction


def trid_hf(xx):
    xx = np.atleast_2d(xx)

    temp1 = np.sum((xx - 1) ** 2, axis=1)
    temp2 = np.sum(xx[:,:-1] * xx[:,1:], axis=1)
    return temp1 - temp2

def adjustable_trid_lf(xx, a4):
    xx = np.atleast_2d(xx)

    temp1 = np.sum((xx - a4) ** 2, axis=1)
    temp2 = np.sum((a4-0.65) * xx[:,:-1] * xx[:,1:] * np.arange(2, 11), axis=1)
    return temp1 - temp2


u_bound = [100]*10
l_bound = [-100]*10


def trid(a4: float):
    """Factory method for adjustable Trid function using parameter value `a4`

    :param a4:  Parameter to tune the correlation between high- and low-fidelity
                functions. Expected values lie in range [0, 1].
    :return:    A MultiFidelityFunction instance
    """

    return MultiFidelityFunction(
        f"adjustable Trid {a4}",
        u_bound, l_bound,
        [trid_hf, partial(adjustable_trid_lf, a4=a4)],
        fidelity_names=['high', 'low'],
    )



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
property_test.py: testing file with property-based tests
for mf2 package
"""

__author__ = 'Sander van Rijn'
__email__ = 's.j.van.rijn@liacs.leidenuniv.nl'


import numpy as np
from hypothesis import given
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import floats, integers, just, tuples
import pytest

from .utils import rescale, ValueRange
import mf2


def quadratic(xx):
    xx = np.atleast_2d(xx)

    return np.sqrt(np.sum(xx**2, axis=1))


simple_square = mf2.MultiFidelityFunction('simple_square', [-1e8], [1e8], [quadratic, quadratic], ['high', 'low'])

# TEST HELPERS -----------------------------------------------------------------
def ndim_array(n):
    return arrays(
        dtype=float,
        shape=tuples(integers(min_value=1, max_value=100), just(n)),
        elements=floats(0, 1),
    )


def _test_1d_array_input(func, x):
    y = func(x)

    assert isinstance(y, np.ndarray)
    assert y.ndim == 1


def _test_2d_array_input(func, x):
    y = func(x)

    assert isinstance(y, np.ndarray)
    assert y.ndim == 1
    assert np.all(np.isfinite(y))

    if len(x) > 1:
        assert np.allclose(func(x[0]), y[0])


def _test_single_function(f, x):
    X = rescale(np.array(x),
                range_in=ValueRange(0, 1),
                range_out=ValueRange(*f.bounds))
    for fidelity in f.functions:
        _test_2d_array_input(fidelity, X.tolist())
        _test_2d_array_input(fidelity, X)  # direct numpy input


# TESTS ------------------------------------------------------------------------

def test_confirm_bifidelity_functions():
    fidelity_names = ['high', 'low']
    for f in mf2.bi_fidelity_functions:
        assert f.fidelity_names is not None
        assert len(f.fidelity_names) == len(f.functions) == 2
        for name in fidelity_names:
            assert name in f.fidelity_dict


@given(integers(min_value=1, max_value=100).flatmap(ndim_array))
@pytest.mark.parametrize("function", [
    simple_square,
    mf2.forrester,
])
def test_nd_functions(function, x):
    _test_single_function(function, x)


@given(integers(1, 100))
@pytest.mark.parametrize("factory", [
    mf2.Forrester,
])
def test_dimensionality_factory_valid_ndim(factory, ndim):
    func = factory(ndim)
    assert func.ndim == ndim


@given(integers(max_value=0))
@pytest.mark.parametrize("factory", [
    mf2.Forrester,
])
def test_dimensionality_factory_invalid_ndim(factory, ndim):
    with pytest.raises(ValueError):
        _ = factory(ndim)


@given(ndim_array(n=2))
def test_double_invert_idempotency(x):
    """Checks that applying invert() twice results in the same values"""
    orig = mf2.branin
    double_inverted = mf2.invert(mf2.invert(orig))
    assert np.allclose(orig.low(x), double_inverted.low(x))
    assert np.allclose(orig.high(x), double_inverted.high(x))


@given(ndim_array(n=2))
def test_invert_currin(x):
    """Test the optimum of the (inverted) Currin function"""
    currin = mf2.currin
    inverted_currin = mf2.invert(mf2.currin)
    x_opt = [0.21666666667, 0]
    y_opt = mf2.currin.high(x_opt)
    inverted_y_opt = y_opt * -1

    assert all(currin.high(x) < y_opt)
    assert all(inverted_currin.high(x) > inverted_y_opt)


@given(ndim_array(n=2))
@pytest.mark.parametrize("function", [
    mf2.bohachevsky,
    mf2.booth,
    mf2.branin,
    mf2.currin,
    mf2.himmelblau,
    mf2.six_hump_camelback,
    mf2.adjustable.branin(0.5),
    mf2.adjustable.paciorek(0.5),
])
def test_2d_functions(function, x):
    _test_single_function(function, x)


@given(ndim_array(n=3))
@pytest.mark.parametrize("function", [
    mf2.adjustable.hartmann3(0.5),
])
def test_3d_functions(function, x):
    _test_single_function(function, x)


@given(ndim_array(n=4))
@pytest.mark.parametrize("function", [
    mf2.park91a,
    mf2.park91b,
])
def test_4d_functions(function, x):
    _test_single_function(function, x)


@given(ndim_array(n=6))
@pytest.mark.parametrize("function", [
    mf2.hartmann6,
])
def test_6d_functions(function, x):
    _test_single_function(function, x)


@given(ndim_array(n=8))
@pytest.mark.parametrize("function", [
    mf2.borehole,
])
def test_8d_functions(function, x):
    _test_single_function(function, x)


@given(ndim_array(n=10))
@pytest.mark.parametrize("function", [
    mf2.adjustable.trid(0.5),
])
def test_10d_functions(function, x):
    _test_single_function(function, x)
